"""Module for user-backend communication."""

from abc import ABCMeta, abstractmethod
from collections import deque

INFO_VISIT, INFO_FINISH, INFO_LEG = range(3)
(
    INPUT_NR_PLAYERS,
    INPUT_START_VALUE,
    INPUT_PLAYER_NAME,
    INPUT_NR_LEGS,
    INPUT_ANOTHER_SESSION,
    INPUT_THROW,
) = range(6)


class CommunicatorBase:
    """Communicators provide a high-level interface to fetch data requested by
    the game routine and to display information about game stastics.
    The only member variables are three callables functioning as input and
    output (info and error) method, resp.
    A dictionary contains prompts for all possible input queries including
    sanitization kwargs. It can be overwritten in subclasses.
    """

    __metaclass__ = ABCMeta

    def __init__(self, input_method, output_info_method, output_error_method=None):
        self._input_method = input_method
        self._output_info_method = output_info_method
        self._output_error_method = output_error_method or output_info_method
        self._input_prompts = {
            INPUT_NR_PLAYERS: {
                "prompt": "Nr of players: ",
                "kwargs": {"type_": int, "min_": 1},
            },
            INPUT_START_VALUE: {
                "prompt": "Start value: ",
                "kwargs": {"type_": int, "min_": 2},
            },
            INPUT_PLAYER_NAME: {
                "prompt": "Name of player {}: ",
                "kwargs": {"min_": 1},
            },
            INPUT_NR_LEGS: {
                "prompt": "Nr of legs: ",
                "kwargs": {"type_": int, "min_": 1},
            },
            INPUT_ANOTHER_SESSION: {
                "prompt": "Again (y/n/q)? ",
                "kwargs": {"choices": "ynq"},
            },
            INPUT_THROW: {
                "prompt": "{}'s score: ",
                "kwargs": {},
            },
        }

    def get_input(self, input_mode, *format_args):
        """Prompt the user to give valid input. If invalid, display error and
        repeat. Prompt parameters are read acc. to the given `input_mode` and,
        if specified, formatted using `format_args`.
        """
        while True:
            prompt_parameters = self._input_prompts[input_mode]
            prompt = prompt_parameters["prompt"].format(*format_args)
            kwargs = prompt_parameters["kwargs"]
            try:
                user_input = self._input_method(prompt)
                return sanitized_input(user_input, **kwargs)
            except SanitizationError as e:
                self.print_error(error=e)

                if isinstance(e, MinLargerMaxError):
                    # re-raise to avoid infinite loop
                    raise

    @abstractmethod
    def print_info(self, message_type, **data):
        """Print some info to the frontend."""

    @abstractmethod
    def print_error(self, **data):
        """Report error to the frontend."""

    def print_newline(self):
        self._output_info_method("")


class TestingCommunicator(CommunicatorBase):
    """Communicator for testing game procedures (visits, legs, sessions).
    Requires sensible input data representing throws or visits.
    """

    def __init__(self, *data):
        self._data = deque((str(d) for d in data))

        def pop(deque_):
            return deque_.popleft()

        super().__init__(pop, lambda _: None)

    def get_input(self, input_mode=INPUT_THROW, *format_args):
        """Pop element from data deque."""
        element = self._input_method(self._data)
        return sanitized_input(element, **self._input_prompts[input_mode]["kwargs"])

    def print_info(self, message_type, **data):
        """Does not do anything."""

    def print_error(self, **data):
        """Does not display any text. Re-raises any exception being passed."""
        raise data["error"]


def sanitized_input(ui, type_=str, min_=None, max_=None, choices=None):
    """Helper method to retrieve sanitized user input.
    Attempts conversion to requested type and validates the input.
    Input that's supposed to be string is stripped from whitespace.
    Inspiration was taken from
    https://stackoverflow.com/questions/23294658/asking-the-user-for-input-until-they-give-a-valid-response

    :param ui: user input that is to be sanitized
    :param type_: Expected type of the input. Defaults to str
    :param min_: Minimum value for input number, or minimum length for input str
    :param max_: Maximum value for input number, or maximum length for input str
    :param choices: Iterable that the input has to be contained in,
        default: None, meaning that no check is performed
    """

    if min_ is not None and max_ is not None and max_ < min_:
        raise MinLargerMaxError("min_ must be less than or equal to max_.")

    if type_ is not None:
        try:
            ui = type_(ui)
        except ValueError:
            raise SanitizationError(f"Input != {type_.__name__}.")
        if type_ is str:
            ui = ui.strip()

    if max_ is not None:
        if type_ is str:
            if len(ui) > max_:
                raise SanitizationError(f"Input length > {max_}")
        else:
            if ui > max_:
                raise SanitizationError(f"Input > {max_}")
    elif min_ is not None:
        if type_ is str:
            if len(ui) < min_:
                raise SanitizationError(f"Input length < {min_}")
        else:
            if ui < min_:
                raise SanitizationError(f"Input < {min_}")

    if choices is not None and (not len(str(ui)) or ui not in choices):
        raise SanitizationError(f"Input != [{','.join(str(c) for c in choices)}]")

    return ui


class SanitizationError(Exception):
    """Raised if sanitization cannot be performed with given parameters."""


class MinLargerMaxError(SanitizationError):
    pass



from .communication import INFO_LEG
from .database import LogEntryBase


class Session(LogEntryBase):
    """Representation of a darts session.
    Initialized with a list of players and the number of legs to win.
    """

    def __init__(self, players, nr_legs=1, log_parent=None, communicator=None):
        super().__init__(log_parent, players=",".join([p.name for p in players]))

        self._players = players
        self._nr_legs = nr_legs
        self._communicator = communicator

    def _player_won_enough_legs(self):
        largest_nr_of_victories = max((p.nr_won_legs for p in self._players))
        return self._nr_legs <= largest_nr_of_victories

    def run(self):
        """Play until a player has won the predefined number of legs."""
        Leg.start_player_index = 0

        for p in self._players:
            p._nr_won_legs = 0

        while not self._player_won_enough_legs():
            leg = Leg(self._players, log_parent=self)
            leg.run()

            self.save()

            self._communicator.print_info(INFO_LEG, players=self._players)


class Leg(LogEntryBase):
    """Representation of a darts leg. Players start alternatingly which is
    established by the static variable `start_player_index` that is incremented
    every time a Leg instance is created.
    """

    start_player_index = 0

    def __init__(self, players, log_parent=None):
        super().__init__(log_parent)

        self._players = players
        self._nr_players = len(self._players)

        self._current_player_index = Leg.start_player_index % self._nr_players
        Leg.start_player_index += 1

    def run(self):
        """Players are taking turns and playing until a player wins."""
        for p in self._players:
            p.reset()

        while True:
            current_player = self._players[self._current_player_index]

            visit = Visit(current_player, log_parent=self)
            visit.run()

            if current_player.victorious():
                current_player.just_won_leg()
                break

            self._current_player_index += 1
            self._current_player_index %= self._nr_players


class Visit(LogEntryBase):
    """Representation of a player's visit. Only used to log visit information to
    the parent leg log.
    """

    def __init__(self, player, log_parent=None):
        super().__init__(log_parent, player=player.name)
        self._player = player

    def run(self):
        """The player plays one visit. The log entry is updated."""
        self._player.play()
        self._log_entry.attrib.update(
            dict(
                points=str(self._player.visit_sum()), throws=str(3 - self._player.darts)
            )
        )



from .communication import INFO_FINISH, INFO_VISIT, INPUT_THROW


class Player:
    """Representation of a player.
    Holds long-term, per-game and per-visit information. Provides routines to
    perform a classic 501.
    Each player has a unique name that are passed at initialization.
    It is also required to pass a start value to initialize the player's
    remaining score."""

    INDEX = 0

    def __init__(self, name, start_value=501, communicator=None):
        self._name = name
        self._index = Player.INDEX
        Player.INDEX += 1
        self._start_value = start_value
        self._nr_won_legs = 0
        self._communicator = communicator

        self._score_left = start_value
        self._throws = 0

        self._darts = 3
        self._visit = []

    def begin(self):
        """Reset actions at the beginning of a player's turn."""
        if self._visit:
            self._visit = []
        self._darts = 3

    def reset(self):
        """Reset actions at the beginning of a leg."""
        self._throws = 0
        self._score_left = self._start_value

    def victorious(self):
        return self._score_left == 0

    def substract(self, score, is_total=True):
        """Substract `score` from the player's `score_left`. If `is_total` is
        true or if the score is obviously the last one of a player's turn, the
        number of remaining darts is set to zero.
        Note that score has to be valid since this method does not do any
        checks on its own."""
        self._score_left -= score
        # TODO this logic should be moved to _process_score
        if is_total or score + sum(self._visit) == 180:
            self._throws += self._darts
            self._darts = 0
        else:
            self._throws += 1
            self._darts -= 1
        self._visit.append(score)

    def score_valid(self, score):
        """Check whether `score` is valid, i.e. the current visit must not
        exceed 180, the player must not be busted and the score has to be
        legitimatly thrown (e.g. one cannot score 130 with two remaining darts.)

        Raises a ValueError if the score is invalid.
        """
        difference = self._score_left - score

        if (
            score + sum(self._visit) > 180
            or score > 60 * self._darts
            or difference < 0
            or difference == 1
        ):
            raise ValueError

        return True

    def print_info(self):
        """Print information on player's left score and darts and, optionally,
        finishing options."""
        self._communicator.print_info(INFO_VISIT, player=self)
        self._communicator.print_info(INFO_FINISH, player=self)

    @property
    def score_left(self):
        return self._score_left

    @property
    def darts(self):
        return self._darts

    @property
    def name(self):
        return self._name

    @property
    def throws(self):
        return self._throws

    def visit_sum(self):
        return sum(self._visit)

    @property
    def nr_won_legs(self):
        return self._nr_won_legs

    def just_won_leg(self):
        self._nr_won_legs += 1

    def play(self):
        """Read score input given by the user while checking for errors.
        Player information is printed before each throw. If _process_score()
        does not raise any errors, the input score is substracted from the
        player's remaining score.
        The procedure is repeated if the player has darts remaining and has not
        yet won.
        """
        self.begin()

        while self._darts and not self.victorious():
            self.print_info()
            input_ = self._communicator.get_input(INPUT_THROW, self._name)
            try:
                score, is_total = self._process_score(input_)
                self.substract(score, is_total)
            except ValueError as e:
                self._communicator.print_error(error=e)
        self._communicator.print_info(INFO_VISIT, player=self)
        self._communicator.print_newline()

    def _process_score(self, score):
        """Parse the passed score. Valid options are:
        - x (an integer between 0 and 180)
        - xd (an integer between 0 and 180, with the character 'd' appended,
          indicating that the player's turn is over ('done'))
        - d (equivalent to '0d')
        - b (the character 'b', indicating that the player busted. The player's
          score will be reset in substract())
        A ValueError is thrown if an invalid score is passed. This can be
        caused by a score that exceeds the remaining score allowed in the
        current visit or by a typo."""
        try:
            score = score.lower()
            if score.endswith("d"):
                if len(score) == 1:
                    score = 0
                else:
                    score = int(score[:-1])
                is_total = True
            elif score == "b":
                # player busted, undo whatever he scored so far in his turn
                # all three darts are taken into account for the average, see
                # http://www.dartsnutz.net/forum/archive/index.php/thread-18910.html
                score = -self.visit_sum()
                is_total = True
            else:
                score = int(score)
                is_total = False

            if self.score_valid(score):
                return score, is_total
        except ValueError:
            # custom message
            raise ValueError(f"Invalid input: {score}")



import json
import os.path
from urllib import request

import lxml.html

contents = (
    request.urlopen("https://www.uni-muenster.de/ZIV.GuidoWessendorf/checkout.html")
    .read()
    .decode("utf-8")
)

finishes = {}
table_end = -1
while True:
    table_start = contents.find("TABLE", table_end + 1)
    if table_start == -1:
        break
    table_end = contents.find("TABLE", table_start + 1)

    # http://stackoverflow.com/questions/20418807/python-parse-html-table-using-lxml
    table = lxml.html.fromstring(contents[table_start - 2 : table_end + 7])  # noqa
    data = table.xpath("//tr/td//text()")

    # skip table header
    i = 25

    # iterate over all table entries. First one can be converted to int (the
    # score). From there, all non-whitespace entries are collected. Integers might
    # appear, indicating a single field. They must not be interpreted as score,
    # hence the j counter is introduced.
    while i < len(data):
        score = int(data[i])
        i += 1
        valid = True
        finish = []
        j = 0
        while valid and i < len(data):
            try:
                if j < 12:
                    raise ValueError
                int(data[i])
                valid = False
            except ValueError:
                if len(data[i].strip()):
                    finish.append(data[i])
                i += 1
                j += 1
        finishes[score] = finish

# clean up
finishes.pop(40)
for score, finish in finishes.items():
    finishes[score] = []
    while len(finish):
        finish_list = finish[:2]
        if len(finish) > 2 and finish[2] != "-":
            finish_list.append(finish[2])
        finish = finish[3:]
        finishes[score].append(finish_list)

dirname = os.path.dirname(os.path.abspath(__file__))
with open(os.path.join(dirname, "finishes.json"), "w") as file:
    json.dump(eval(finishes.__repr__()), file)



import unittest
from collections import Counter
from unittest import mock
from xml.etree import ElementTree as etree

from pydartz.communication import TestingCommunicator
from pydartz.database import LogEntryBase, PlayerEntry, analyze_sessions
from pydartz.player import Player
from pydartz.session import Session, Visit


class PlayerEntryTestCase(unittest.TestCase):
    def test_default_init(self):
        entry = PlayerEntry()
        self.assertEqual(0, entry._throws)
        self.assertListEqual([], entry._points)
        self.assertEqual(0, len(entry._finishes))

    def test_init_with_stats(self):
        player_stats = dict(
            throws=123, points=4567, finishes=Counter({20: 89}), darters=Counter()
        )
        entry = PlayerEntry(player_stats=player_stats)
        self.assertEqual(123, entry._throws)
        self.assertEqual(4567, entry._points)
        self.assertEqual(89, entry._finishes[20])
        self.assertEqual(0, len(entry._darters))

    def test_update(self):
        entry = PlayerEntry()
        entry.update(9, 501)
        self.assertEqual(9, entry._throws)
        self.assertEqual(501, entry._points[0])
        self.assertEqual(0, len(entry._finishes))
        self.assertEqual(501, entry.total_points())

    def test_to_dict(self):
        player_stats = dict(
            throws=123, points=4567, finishes=Counter({20: 89}), darters=Counter({9: 2})
        )
        entry = PlayerEntry("Michael", player_stats)
        self.assertDictEqual(player_stats, entry.to_dict()["Michael"])

    def test_average(self):
        entry = PlayerEntry()
        entry.update(10, 420)
        self.assertEqual(42, entry.average())


class LogEntryBaseTestCase(unittest.TestCase):
    def test_save(self):
        parent = LogEntryBase()
        parent.save = mock.MagicMock()
        log_entry_base = LogEntryBase(parent=parent)
        log_entry_base.save()
        self.assertEqual(parent.save.call_count, 1)


class AnalysisTestCase(unittest.TestCase):
    def test_analyse_sessions(self):
        sessions = etree.Element("sessions")
        communicator = TestingCommunicator("180d", 60, 60, 57, 60, 60, 24)
        session = Session(
            [Player("Peter", communicator=communicator)],
            1,
            log_parent=sessions,
            communicator=communicator,
        )
        session.run()

        player_entry = analyze_sessions(sessions)["Peter"]

        self.assertEqual(player_entry.total_points(), 501)
        self.assertAlmostEqual(player_entry.average(), 501 / 9)
        self.assertEqual(player_entry.throws, 9)
        self.assertEqual(player_entry._finishes[144], 1)
        self.assertEqual(player_entry._darters[9], 1)


class PlayerLoggingTestCase(unittest.TestCase):
    def setUp(self):
        self.player = Player("Raymond")

    def test_single_visit_logging(self):
        log_entry = etree.Element("leg")
        self.player._communicator = TestingCommunicator("60", "50", "40")
        visit = Visit(self.player, log_parent=log_entry)
        visit.run()
        self.assertEqual(len(log_entry), 1)
        visit_log_entry = log_entry[0]
        self.assertEqual(visit_log_entry.get("player"), self.player.name)
        self.assertEqual(visit_log_entry.get("points"), "150")
        self.assertEqual(visit_log_entry.get("throws"), "3")


if __name__ == "__main__":
    unittest.main()



import argparse
import os
import shutil
import sys
import time

from . import __version__
from .communication import INFO_FINISH, INFO_LEG, INFO_VISIT, CommunicatorBase
from .database import Sessions, analyze_sessions
from .finishes import FINISHES
from .game import Game

log_dir = os.path.expanduser("~/.local/share/pydartz")
os.makedirs(log_dir, exist_ok=True)
log_filepath = os.path.join(log_dir, "stats.xml")
with open(log_filepath, "a"):
    pass
sessions_log = Sessions(log_filepath=log_filepath)


def main():
    args = vars(_parse_command())
    player_names = args.pop("stats", None)

    if player_names is not None:
        player_entries = analyze_sessions(sessions_log._log_entry)
        for name, entry in player_entries.items():
            if len(player_names) and name not in player_names:
                continue
            print(entry.information())
        sys.exit(0)

    time.sleep(1)
    os.system("cls || clear")
    _display_banner()

    g = Game(CliCommunicator(), sessions_log)
    try:
        g.run()
        _play_ending_song()
    except KeyboardInterrupt:
        print("\n\nUser quit by pressing Ctrl-C. Good bye!")


class CliCommunicator(CommunicatorBase):
    """Command Line Interface communicator that request user input using
    Python's builtin 'input' method and that prints to stdout.
    """

    def __init__(self):
        super().__init__(input, print)

    def print_info(self, message_type, **data):
        output = None

        if message_type == INFO_VISIT:
            player = data["player"]
            info = ["no darts", "one dart", "two darts", "three darts"][player.darts]
            output = f"{player.name} has {player.score_left} and {info} left."
        elif message_type == INFO_FINISH:
            player = data["player"]
            score_left = str(player.score_left)
            if score_left in FINISHES:
                output = "Finish options:"
                for finish in FINISHES[score_left]:
                    output += "\n\t" + " ".join(finish)
        elif message_type == INFO_LEG:
            players = data["players"]
            output = "\n".join(f"    {p.name}: {p.nr_won_legs:2d}" for p in players)
            output += "\n"
            output += 80 * "="

        if output is not None:
            self._output_info_method(output)

    def print_error(self, **data):
        output = str(data["error"])
        self._output_error_method(output)


def _parse_command():
    parser = argparse.ArgumentParser()
    parser.suggest_on_error = True

    parser.add_argument(
        "-s", "--stats", metavar="NAME", nargs="*", help="display player stats"
    )
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=f"pydartz version {__version__}",
        help="display version info and exit",
    )  # pragma: no cover

    return parser.parse_args()


def _display_banner():
    terminal_width = shutil.get_terminal_size((80, 20)).columns
    if terminal_width < 94:
        # Display smaller banner for narrow terminals
        print(" pydarts ".center(terminal_width, "#"))
        return

    # fmt: off
    print(r"      ___                                   ___           ___                         ___     ")
    print(r"     /\  \                   _____         /\  \         /\  \                       /\__\    ")
    print(r"    /::\  \       ___       /::\  \       /::\  \       /::\  \         ___         /:/ _/_   ")
    print(r"   /:/\:\__\     /|  |     /:/\:\  \     /:/\:\  \     /:/\:\__\       /\__\       /:/ /\  \  ")
    print(r"  /:/ /:/  /    |:|  |    /:/  \:\__\   /:/ /::\  \   /:/ /:/  /      /:/  /      /:/ /::\  \ ")
    print(r" /:/_/:/  /     |:|  |   /:/__/ \:|__| /:/_/:/\:\__\ /:/_/:/__/___   /:/__/      /:/_/:/\:\__\ ")
    print(r" \:\/:/  /    __|:|__|   \:\  \ /:/  / \:\/:/  \/__/ \:\/:::::/  /  /::\  \      \:\/:/ /:/  /")
    print(r"  \::/__/    /::::\  \    \:\  /:/  /   \::/__/       \::/~~/~~~~  /:/\:\  \      \::/ /:/  / ")
    print(r"   \:\  \    ~~~~\:\  \    \:\/:/  /     \:\  \        \:\~~\      \/__\:\  \      \/_/:/  /  ")
    print(r"    \:\__\        \:\__\    \::/  /       \:\__\        \:\__\          \:\__\       /:/  /   ")
    print(r"     \/__/         \/__/     \/__/         \/__/         \/__/           \/__/       \/__/    ")
    print()
    # fmt: on


def _play_ending_song():
    data_directory = "data"
    filename = "chase_the_sun.wav"
    try:
        import simpleaudio

        wave_obj = simpleaudio.WaveObject.from_wave_file(
            os.path.join(sys.prefix, data_directory, filename)
        )
        play_obj = wave_obj.play()
        play_obj.wait_done()
    except ImportError:
        try:
            import subprocess

            media_player = os.getenv("PYDARTZ_MEDIA_PLAYER", "mpv")
            subprocess.run(
                [media_player, f"{data_directory}/{filename}"],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )
        except Exception:
            print("Run 'pip install -U pydartz[audio]' for sound support!")



import contextlib
import os.path
import time
from abc import ABCMeta
from collections import Counter
from xml.etree import ElementTree as etree


class LogEntryBase:
    """Abstract base class for all classes that are supposed to support
    logging.

    It holds a member `_log_entry` of type `xml.etree._Element`. Its tag is
    derived from the lowercase class name. If a `parent` argument (an instance
    of a class that implements an `append` method, such as a list or this
    class) is passed, the `log_entry` is appended, s.t. the hierarchy of the
    `session` classes is reflected in the XML log.
    The initialization time of each instance is stored as timestamp for later
    reference.
    Any additional kwargs are passed to `xml.etree.Element` where they are
    interpreted as XML attributes.
    """

    __metaclass__ = ABCMeta

    DT_FORMAT = "%Y%m%d-%H%M%S"

    def __init__(self, parent=None, **kwargs):
        tag = self.__class__.__name__.lower()
        self._log_entry = etree.Element(
            tag, timestamp=time.strftime(self.DT_FORMAT), **kwargs
        )

        self._parent = parent
        if self._parent is not None:
            self._parent.append(self._log_entry)

    def append(self, child_log_entry):
        self._log_entry.append(child_log_entry)

    def save(self):
        with contextlib.suppress(AttributeError):
            # parent is e.g. None or a list
            self._parent.save()


class Sessions(LogEntryBase):
    """Wrapper around the top-most XML-log element 'sessions'. Does not have a
    parent.
    """

    def __init__(self, log_filepath=None, **kwargs):
        """Parse existing tree from filepath or create new sessions element."""
        if (
            log_filepath is not None
            and os.path.exists(log_filepath)
            and os.path.getsize(log_filepath) > 0
        ):
            self._log_entry = etree.parse(log_filepath).getroot()
        else:
            self._log_entry = etree.Element("sessions")
        self._log_filepath = log_filepath

    def save(self):
        """Write the `log_entry` XML object to disk."""
        if self._log_filepath is not None:
            tree = etree.ElementTree(self._log_entry)
            tree.write(self._log_filepath, xml_declaration=True, encoding="utf-8")


class PlayerEntry:
    """A container to facilitate evaluation of player statistics.

    It stores the player's total throws, points (visit per visit for e.g.
    histogram evaluation), finishes and darters.
    Some basic update and query methods are implemented.
    """

    def __init__(self, name=None, player_stats=None):
        self._name = name
        self._throws = 0
        self._points = []
        self._finishes = Counter()
        self._darters = Counter()

        if player_stats:
            self._throws = player_stats["throws"]
            self._points = player_stats["points"]
            self._finishes = player_stats["finishes"]
            self._darters = player_stats["darters"]

    def update(self, throws=0, points=0, darter=None):
        self._throws += throws
        self._points.append(points)

        # only called if current player won. Points of last leg are passed
        if darter is not None:
            self._finishes[points] += 1
            self._darters[darter] += 1

    def update_from_log(self, log_element):
        """Convenience method to update from a `xml.etree._Element` visit."""
        self.update(
            points=int(log_element.get("points")), throws=int(log_element.get("throws"))
        )

    def to_dict(self):
        """Returns a dict representation of the PlayerEntry data."""
        return {
            self._name: dict(
                throws=self._throws,
                points=self._points,
                finishes=self._finishes,
                darters=self._darters,
            )
        }

    def average(self):
        if self._throws != 0:
            return self.total_points() / self._throws
        return -1

    def total_points(self):
        return sum(self._points)

    @property
    def throws(self):
        return self._throws

    def information(self):
        info = [
            self._name + ":",
            f"Legs won: {sum(self._finishes.values())}",
            f"Average: {3 * self.average():.2f}",
            f"Highscore: {max(self._points):3d}",
            "Finishes:",
        ]
        for finish in sorted(self._finishes)[::-1]:
            info.append(f"    {finish:3d}: {self._finishes[finish]}")
        info.append("Darters:")
        for darter in sorted(self._darters):
            info.append(f"    {darter:3d}-darter: {self._darters[darter]}")

        return "\n".join(info)


def analyze_sessions(sessions):
    """Analyze a `xml.etree._Element` sessions object.

    First all player names occurring are collected and a dict of PlayerEntrys
    is built.
    It is iterated over every session played so far. For every visit, the
    corresponding PlayerEntry is updated. The data of the last visit of a leg
    is used to update the winner's finishes and darters.
    """
    player_names = set()
    for session in sessions:
        for name in session.get("players").split(","):
            player_names.add(name)

    players = {name: PlayerEntry(name) for name in player_names}

    for session in sessions:
        for leg in session:
            winner_name = leg[-1].get("player")
            winner = players[winner_name]
            winner_old_throws = winner.throws

            for visit in leg[:-1]:
                name = visit.get("player")
                player_entry = players[name]
                player_entry.update_from_log(visit)

            final_throws = int(leg[-1].get("throws"))
            final_points = int(leg[-1].get("points"))
            # X-darter per player
            visit_throws = final_throws + winner.throws - winner_old_throws
            winner.update(points=final_points, throws=final_throws, darter=visit_throws)

    return players



import unittest
from datetime import datetime
from unittest import mock

from pydartz.communication import TestingCommunicator
from pydartz.player import Player
from pydartz.session import Leg, Session


def _180_single_player_communicator():
    return TestingCommunicator("180d", 60, 60, 57, 60, 60, 24)


class LegTestCase(unittest.TestCase):
    def test_single_player_9_darter(self):
        communicator = _180_single_player_communicator()
        mike = Player("Mike", communicator=communicator)
        leg = Leg([mike])
        leg.run()
        self.assertEqual(leg._current_player_index, 0)
        self.assertTrue(mike.victorious())

    def test_two_player_101(self):
        communicator = TestingCommunicator(
            # Hans   Fritz
            "60d",
            19,
            17,
            3,
            19,
            "b",
            12,
            50,
        )
        hans = Player("Hans", 101, communicator=communicator)
        fritz = Player("Fritz", 101, communicator=communicator)
        # static variable is increased by previous tests...
        # in reality, the Session handles this
        Leg.start_player_index = 0
        leg = Leg([hans, fritz])
        leg.run()

        self.assertEqual(leg._current_player_index, 1)
        self.assertTrue(fritz.victorious())
        self.assertEqual(hans.score_left, 41)

    def test_logging_without_parent(self):
        leg = Leg(["Peter"])
        # might fail around midnight...
        self.assertEqual(
            datetime.strptime(leg._log_entry.get("timestamp"), Leg.DT_FORMAT).day,
            datetime.today().day,
        )

    def test_logging_with_parent(self):
        log_parent = []
        Leg(["Mike"], log_parent=log_parent)
        log_entry = log_parent[0]
        self.assertEqual(log_entry.tag, "leg")
        self.assertEqual(len(log_entry), 0)

    def test_single_player_9_darter_logging(self):
        communicator = _180_single_player_communicator()
        players = [Player("Mike", communicator=communicator)]
        log_parent = []
        leg = Leg(players, log_parent=log_parent)
        leg.run()

        visits_log_entry = log_parent[0]
        self.assertEqual(len(visits_log_entry), 3)
        self.assertEqual(visits_log_entry[0].get("throws"), "3")
        self.assertEqual(visits_log_entry[2].get("points"), "144")
        # assert average etc

    def test_single_player_invalid_visit(self):
        with mock.patch.object(TestingCommunicator, "print_error") as print_error_patch:
            communicator = TestingCommunicator(
                "180d", 60, 60, 57, 60, 60, 60, 24  # <-- this is invalid
            )
            player = Player("Mike", communicator=communicator)
            leg = Leg([player])
            leg.run()
        self.assertEqual(print_error_patch.call_count, 1)
        self.assertTrue(player.victorious())


class SessionTestCase(unittest.TestCase):
    def test_single_player_9_darter_session(self):
        communicator = _180_single_player_communicator()
        players = [Player("Peter", communicator=communicator)]
        session = Session(players, 1, communicator=communicator)
        session.run()

        self.assertEqual(len(session._log_entry[0]), 3)
        self.assertEqual(len(session._log_entry), 1)

    def test_two_player_session(self):
        communicator = TestingCommunicator(
            # Adam     Eve
            "60d",
            20,
            11,
            20,
            "40",
            "60d",
            20,
            11,
            20,
            40,
            "60d",
            20,
            11,
            20,
            "40",
        )
        adam = Player("Adam", 100, communicator=communicator)
        eve = Player("Eve", 100, communicator=communicator)
        session = Session([adam, eve], 2, communicator=communicator)
        session.run()

        self.assertTrue(adam.victorious())
        self.assertEqual(eve.score_left, 49)


if __name__ == "__main__":
    unittest.main()



import unittest

from pydartz.communication import TestingCommunicator
from pydartz.player import Player


class PlayerEntryTestCase(unittest.TestCase):
    def setUp(self):
        self.player = Player("Raymond")

    def test_not_victorious(self):
        self.assertFalse(self.player.victorious())
        self.assertEqual(self.player.nr_won_legs, 0)

    def test_scores_valid(self):
        self.assertTrue(self.player.score_valid(180))
        self.player._visit.append(60)
        self.assertRaises(ValueError, self.player.score_valid, 121)
        # hack low remaining score
        self.player._score_left = 40
        self.assertTrue(self.player.score_valid(20))
        # busted
        self.assertRaises(ValueError, self.player.score_valid, 39)
        self.assertRaises(ValueError, self.player.score_valid, 42)
        # finish
        self.assertTrue(self.player.score_valid(40))

    def test_substract(self):
        # first visit
        self.player.substract(180, True)
        self.assertEqual(self.player.score_left, 321)
        self.assertEqual(self.player.throws, 3)
        self.assertEqual(self.player.darts, 0)
        self.assertEqual(self.player.visit_sum(), 180)
        # second visit
        self.player.begin()
        self.assertEqual(self.player.visit_sum(), 0)
        self.assertEqual(self.player.darts, 3)
        self.player.substract(60, False)
        self.assertEqual(self.player.throws, 4)
        self.assertEqual(self.player.darts, 2)
        self.player.substract(60, False)
        self.player.substract(60, False)
        self.assertEqual(self.player.throws, 6)
        self.assertEqual(self.player.darts, 0)
        self.assertEqual(self.player.visit_sum(), 180)
        self.assertEqual(self.player.score_left, 141)
        # third visit, busting for testing reasons
        self.player.begin()
        self.player.substract(0, True)
        self.assertEqual(self.player.throws, 9)
        self.assertEqual(self.player.score_left, 141)
        # fourth visit, busts again
        self.player.begin()
        self.player.substract(60, False)
        self.assertEqual(self.player.score_left, 81)
        self.player.substract(-60, True)
        self.assertEqual(self.player.throws, 12)
        self.assertEqual(self.player.score_left, 141)
        # fifth visit, finishes
        self.player.begin()
        self.player.substract(141, True)
        self.assertEqual(self.player.throws, 15)
        self.assertTrue(self.player.victorious())

    def test_process_score(self):
        self.assertTupleEqual(self.player._process_score("99d"), (99, True))
        self.assertTupleEqual(self.player._process_score("60"), (60, False))
        # bust (score is actually impossible, returns 0 because substract() not called)
        self.assertTupleEqual(self.player._process_score("b"), (0, True))
        self.assertRaises(ValueError, self.player._process_score, "200")
        self.assertRaises(ValueError, self.player._process_score, "hi")

    def test_play(self):
        communicator = TestingCommunicator(
            60,
            60,
            60,
            180,
            1,
            20,
            "d",
            "20d",
            60,
            60,  # invalid
            "darts<3",  # invalid
            50,
            "50",
        )
        self.player._communicator = communicator
        self.player.play()
        self.assertEqual(self.player.score_left, 321)
        self.assertEqual(self.player.darts, 0)
        self.player.play()
        self.assertEqual(self.player.score_left, 141)
        self.assertEqual(self.player.darts, 0)
        self.player.play()
        self.assertEqual(self.player.score_left, 120)
        self.assertEqual(self.player.darts, 0)
        self.player.play()
        self.assertEqual(self.player.score_left, 100)
        self.assertEqual(self.player.darts, 0)
        self.assertRaises(ValueError, self.player.play)
        self.assertEqual(self.player.score_left, 40)
        self.assertEqual(self.player.darts, 2)
        # FIXME bypass begin() when calling play(), because it would log the
        # first 60 abcve
        score, is_total = self.player._process_score("b")
        self.player.substract(score, is_total)
        self.assertEqual(self.player.score_left, 100)
        self.assertEqual(self.player.darts, 0)
        self.assertRaises(ValueError, self.player.play)
        self.assertEqual(self.player.darts, 3)
        self.player.play()
        self.assertTrue(self.player.victorious())

    def test_reset(self):
        self.player._communicator = TestingCommunicator("50", "111d")
        self.player.play()
        self.player.reset()
        self.assertEqual(self.player.score_left, self.player._start_value)
        self.assertEqual(self.player.throws, 0)

    def test_won_legs(self):
        self.player.just_won_leg()
        self.assertEqual(self.player.nr_won_legs, 1)


if __name__ == "__main__":
    unittest.main()



"""
Module containing top game logic and routine.
"""

from .communication import (
    INPUT_ANOTHER_SESSION,
    INPUT_NR_LEGS,
    INPUT_NR_PLAYERS,
    INPUT_PLAYER_NAME,
    INPUT_START_VALUE,
)
from .player import Player
from .session import Session


class Game:
    def __init__(self, communicator, sessions_log=None):
        self._communicator = communicator
        self._sessions_log = sessions_log
        self._current_session = None

    def run(self):
        """Repeatly creates and runs sessions. Stops if the player decides to
        quit.
        """

        while True:
            if self._current_session is None:
                self._init_session()
            else:
                if not self._query_another_session():  # quit
                    return

            self._current_session.run()

    def _init_session(self):
        """Initializes a new session by querying required input (number of
        players, start value, player names, number of legs to win).
        """

        nr_players = self._communicator.get_input(INPUT_NR_PLAYERS)

        start_value = self._communicator.get_input(INPUT_START_VALUE)

        players = []
        for i in range(nr_players):
            name = self._communicator.get_input(INPUT_PLAYER_NAME, i + 1)
            players.append(Player(name, start_value, communicator=self._communicator))

        nr_legs = self._communicator.get_input(INPUT_NR_LEGS)

        self._current_session = Session(
            players,
            nr_legs,
            log_parent=self._sessions_log,
            communicator=self._communicator,
        )

    def _query_another_session(self):
        """Query the user about how to proceed. Three replies are possible:
        - n (no): play again but ask for new session parameters
        - y (yes): play again, using the previous session's parameters
        - q (quit): quit the game

        :return: False if user requested to quit
        """

        continuing = True
        reply = self._communicator.get_input(INPUT_ANOTHER_SESSION)

        if reply == "n":
            # query parameters for new session
            self._init_session()
        elif reply == "y":
            # copy previous session's parameters
            self._current_session = Session(
                self._current_session._players,
                self._current_session._nr_legs,
                log_parent=self._sessions_log,
                communicator=self._communicator,
            )
        else:
            continuing = False

        return continuing



import unittest

from pydartz.communication import TestingCommunicator
from pydartz.database import analyze_sessions
from pydartz.game import Game


def run_game(communicator):
    log = []
    game = Game(communicator, sessions_log=log)
    game.run()
    return analyze_sessions(log)


class GameTestCase(unittest.TestCase):
    def test_single_player_game(self):
        communicator = TestingCommunicator(
            1,  # nr of players
            501,  # start value
            "Herbert",  # name
            1,  # nr of legs to win
            59,
            59,
            59,  # visit data...
            58,
            "60d",
            1,
            2,
            3,
            "150d",
            50,
            "q",
        )
        infos = run_game(communicator)
        player_info = infos["Herbert"].to_dict()["Herbert"]
        self.assertListEqual(player_info["points"], [177, 118, 6, 150, 50])

    def test_single_player_two_sessions_game(self):
        communicator = TestingCommunicator(
            1,
            16,
            "Anton",
            2,
            16,
            8,
            8,
            "y",
            4,
            4,
            8,
            16,
            "q",
        )
        infos = run_game(communicator)
        player_info = infos["Anton"].to_dict()["Anton"]
        self.assertEqual(player_info["throws"], 7)
        self.assertListEqual(player_info["points"], [16 for _ in range(4)])

    def test_two_various_sessions_game(self):
        communicator = TestingCommunicator(
            1,
            16,
            "Anton",
            2,
            8,
            8,
            8,
            8,
            "n",
            2,
            32,
            "Anton",
            "Herbert",
            1,
            16,
            16,
            "q",
        )
        infos = run_game(communicator)
        player_info = infos["Anton"].to_dict()["Anton"]
        self.assertEqual(player_info["throws"], 6)
        self.assertListEqual(player_info["points"], [16, 16, 32])
        self.assertEqual(player_info["darters"][2], 3)


if __name__ == "__main__":
    unittest.main()



import unittest

from pydartz.communication import (
    INPUT_ANOTHER_SESSION,
    INPUT_NR_LEGS,
    INPUT_NR_PLAYERS,
    INPUT_PLAYER_NAME,
    INPUT_START_VALUE,
    INPUT_THROW,
    MinLargerMaxError,
    SanitizationError,
    TestingCommunicator,
    sanitized_input,
)


class SanitizedInputTestCase(unittest.TestCase):
    def test_min_larger_max(self):
        self.assertRaises(MinLargerMaxError, sanitized_input, "foo", min_=1, max_=0)

    def test_wrong_type(self):
        self.assertRaises(SanitizationError, sanitized_input, "foo", type_=int)

    def test_too_large(self):
        self.assertRaises(SanitizationError, sanitized_input, "42", type_=int, max_=11)

    def test_too_small(self):
        self.assertRaises(
            SanitizationError, sanitized_input, "42", type_=int, min_=1337
        )

    def test_too_short(self):
        self.assertRaises(SanitizationError, sanitized_input, "bar", min_=5)

    def test_valid_str(self):
        word = "foo"
        self.assertEqual(sanitized_input(word, min_=1, max_=3), word)

    def test_valid_int(self):
        self.assertEqual(sanitized_input("123", type_=int, max_=180), 123)

    def test_strip(self):
        self.assertEqual(sanitized_input(" 99  ", type_=str), "99")

    def test_valid_choices(self):
        self.assertEqual(sanitized_input("a", choices="abc"), "a")

    def test_invalid_choices(self):
        self.assertRaises(SanitizationError, sanitized_input, "Y", choices="yn")

    def test_invalid_int_choices(self):
        self.assertRaises(
            SanitizationError, sanitized_input, "4", choices=list(range(3)), type_=int
        )

    def test_too_long(self):
        self.assertRaises(SanitizationError, sanitized_input, "bar", max_=2)

    def test_empty_string_not_in_choices(self):
        self.assertRaises(SanitizationError, sanitized_input, "", choices="ab")


class TestingCommunicatorTestCase(unittest.TestCase):
    def setUp(self):
        self.data = (11, "2d", 180, "1", "d", "Pete", "n")
        self.communicator = TestingCommunicator(*self.data)

    def test_data(self):
        for i, d in enumerate(self.data):
            self.assertEqual(str(d), self.communicator._data[i])

    def test_get_input(self):
        self.assertEqual(str(self.data[0]), self.communicator.get_input())
        self.assertEqual(self.data[1], self.communicator.get_input())
        self.assertEqual(str(self.data[2]), self.communicator.get_input())
        self.assertEqual(
            int(self.data[3]), self.communicator.get_input(INPUT_NR_PLAYERS)
        )
        self.assertEqual(
            self.data[4], self.communicator.get_input(INPUT_THROW, "Simon")
        )
        self.assertEqual(
            self.data[5], self.communicator.get_input(INPUT_PLAYER_NAME, (1,))
        )
        self.assertEqual(
            self.data[6], self.communicator.get_input(INPUT_ANOTHER_SESSION)
        )

    def test_print_info(self):
        self.assertIsNone(self.communicator.print_info("some text"))

    def test_print_error(self):
        class TestException(Exception):
            pass

        self.assertRaises(
            TestException, self.communicator.print_error, error=TestException()
        )

    def test_get_typed_input(self):
        self.assertEqual(str(self.data[0]), self.communicator.get_input())
        self.assertEqual(self.data[1], self.communicator.get_input())
        self.assertEqual(self.data[2], self.communicator.get_input(INPUT_START_VALUE))
        self.assertEqual(int(self.data[3]), self.communicator.get_input(INPUT_NR_LEGS))


if __name__ == "__main__":
    unittest.main()
